@@ -21,10 +21,9 @@ import threading
 import time
 import uuid
 from collections import deque
-from contextlib import contextmanager
 
 import sentry_sdk
-from sentry_sdk._compat import PY33
+from sentry_sdk._compat import PY33, PY311
 from sentry_sdk._types import MYPY
 from sentry_sdk.utils import (
     filename_for_module,
@@ -39,15 +38,15 @@ if MYPY:
     from typing import Callable
     from typing import Deque
     from typing import Dict
-    from typing import Generator
     from typing import List
     from typing import Optional
     from typing import Set
     from typing import Sequence
     from typing import Tuple
     from typing_extensions import TypedDict
-    import sentry_sdk.scope
+
     import sentry_sdk.tracing
+    from sentry_sdk._types import SamplingContext
 
     ThreadId = str
 
@@ -104,39 +103,93 @@ if MYPY:
         },
     )
 
+    ProfileContext = TypedDict(
+        "ProfileContext",
+        {"profile_id": str},
+    )
+
+
+try:
+    from gevent import get_hub as get_gevent_hub  # type: ignore
+    from gevent.monkey import get_original, is_module_patched  # type: ignore
+
+    thread_sleep = get_original("time", "sleep")
+except ImportError:
+
+    def get_gevent_hub():
+        # type: () -> Any
+        return None
+
+    thread_sleep = time.sleep
+
+    def is_module_patched(*args, **kwargs):
+        # type: (*Any, **Any) -> bool
+        # unable to import from gevent means no modules have been patched
+        return False
+
+
+def is_gevent():
+    # type: () -> bool
+    return is_module_patched("threading") or is_module_patched("_thread")
+
 
 _scheduler = None  # type: Optional[Scheduler]
 
+# The default sampling frequency to use. This is set at 101 in order to
+# mitigate the effects of lockstep sampling.
+DEFAULT_SAMPLING_FREQUENCY = 101
+
+
+# The minimum number of unique samples that must exist in a profile to be
+# considered valid.
+PROFILE_MINIMUM_SAMPLES = 2
+
 
 def setup_profiler(options):
-    # type: (Dict[str, Any]) -> None
-
-    """
-    `buffer_secs` determines the max time a sample will be buffered for
-    `frequency` determines the number of samples to take per second (Hz)
-    """
-
+    # type: (Dict[str, Any]) -> bool
     global _scheduler
 
     if _scheduler is not None:
         logger.debug("profiling is already setup")
-        return
+        return False
 
     if not PY33:
         logger.warn("profiling is only supported on Python >= 3.3")
-        return
+        return False
 
-    frequency = 101
+    frequency = DEFAULT_SAMPLING_FREQUENCY
 
-    profiler_mode = options["_experiments"].get("profiler_mode", SleepScheduler.mode)
-    if profiler_mode == SleepScheduler.mode:
-        _scheduler = SleepScheduler(frequency=frequency)
+    if is_gevent():
+        # If gevent has patched the threading modules then we cannot rely on
+        # them to spawn a native thread for sampling.
+        # Instead we default to the GeventScheduler which is capable of
+        # spawning native threads within gevent.
+        default_profiler_mode = GeventScheduler.mode
+    else:
+        default_profiler_mode = ThreadScheduler.mode
+
+    profiler_mode = options["_experiments"].get("profiler_mode", default_profiler_mode)
+
+    if (
+        profiler_mode == ThreadScheduler.mode
+        # for legacy reasons, we'll keep supporting sleep mode for this scheduler
+        or profiler_mode == "sleep"
+    ):
+        _scheduler = ThreadScheduler(frequency=frequency)
+    elif profiler_mode == GeventScheduler.mode:
+        try:
+            _scheduler = GeventScheduler(frequency=frequency)
+        except ImportError:
+            raise ValueError("Profiler mode: {} is not available".format(profiler_mode))
     else:
         raise ValueError("Unknown profiler mode: {}".format(profiler_mode))
+
     _scheduler.setup()
 
     atexit.register(teardown_profiler)
 
+    return True
+
 
 def teardown_profiler():
     # type: () -> None
@@ -241,69 +294,135 @@ def extract_frame(frame, cwd):
     )
 
 
-def get_frame_name(frame):
-    # type: (FrameType) -> str
+if PY311:
 
-    # in 3.11+, there is a frame.f_code.co_qualname that
-    # we should consider using instead where possible
+    def get_frame_name(frame):
+        # type: (FrameType) -> str
+        return frame.f_code.co_qualname  # type: ignore
 
-    f_code = frame.f_code
-    co_varnames = f_code.co_varnames
+else:
 
-    # co_name only contains the frame name.  If the frame was a method,
-    # the class name will NOT be included.
-    name = f_code.co_name
+    def get_frame_name(frame):
+        # type: (FrameType) -> str
 
-    # if it was a method, we can get the class name by inspecting
-    # the f_locals for the `self` argument
-    try:
-        if (
-            # the co_varnames start with the frame's positional arguments
-            # and we expect the first to be `self` if its an instance method
-            co_varnames
-            and co_varnames[0] == "self"
-            and "self" in frame.f_locals
-        ):
-            for cls in frame.f_locals["self"].__class__.__mro__:
-                if name in cls.__dict__:
-                    return "{}.{}".format(cls.__name__, name)
-    except AttributeError:
-        pass
+        f_code = frame.f_code
+        co_varnames = f_code.co_varnames
 
-    # if it was a class method, (decorated with `@classmethod`)
-    # we can get the class name by inspecting the f_locals for the `cls` argument
-    try:
-        if (
-            # the co_varnames start with the frame's positional arguments
-            # and we expect the first to be `cls` if its a class method
-            co_varnames
-            and co_varnames[0] == "cls"
-            and "cls" in frame.f_locals
-        ):
-            for cls in frame.f_locals["cls"].__mro__:
-                if name in cls.__dict__:
-                    return "{}.{}".format(cls.__name__, name)
-    except AttributeError:
-        pass
+        # co_name only contains the frame name.  If the frame was a method,
+        # the class name will NOT be included.
+        name = f_code.co_name
 
-    # nothing we can do if it is a staticmethod (decorated with @staticmethod)
+        # if it was a method, we can get the class name by inspecting
+        # the f_locals for the `self` argument
+        try:
+            if (
+                # the co_varnames start with the frame's positional arguments
+                # and we expect the first to be `self` if its an instance method
+                co_varnames
+                and co_varnames[0] == "self"
+                and "self" in frame.f_locals
+            ):
+                for cls in frame.f_locals["self"].__class__.__mro__:
+                    if name in cls.__dict__:
+                        return "{}.{}".format(cls.__name__, name)
+        except AttributeError:
+            pass
 
-    # we've done all we can, time to give up and return what we have
-    return name
+        # if it was a class method, (decorated with `@classmethod`)
+        # we can get the class name by inspecting the f_locals for the `cls` argument
+        try:
+            if (
+                # the co_varnames start with the frame's positional arguments
+                # and we expect the first to be `cls` if its a class method
+                co_varnames
+                and co_varnames[0] == "cls"
+                and "cls" in frame.f_locals
+            ):
+                for cls in frame.f_locals["cls"].__mro__:
+                    if name in cls.__dict__:
+                        return "{}.{}".format(cls.__name__, name)
+        except AttributeError:
+            pass
+
+        # nothing we can do if it is a staticmethod (decorated with @staticmethod)
+
+        # we've done all we can, time to give up and return what we have
+        return name
 
 
 MAX_PROFILE_DURATION_NS = int(3e10)  # 30 seconds
 
 
+def get_current_thread_id(thread=None):
+    # type: (Optional[threading.Thread]) -> Optional[int]
+    """
+    Try to get the id of the current thread, with various fall backs.
+    """
+
+    # if a thread is specified, that takes priority
+    if thread is not None:
+        try:
+            thread_id = thread.ident
+            if thread_id is not None:
+                return thread_id
+        except AttributeError:
+            pass
+
+    # if the app is using gevent, we should look at the gevent hub first
+    # as the id there differs from what the threading module reports
+    if is_gevent():
+        gevent_hub = get_gevent_hub()
+        if gevent_hub is not None:
+            try:
+                # this is undocumented, so wrap it in try except to be safe
+                return gevent_hub.thread_ident
+            except AttributeError:
+                pass
+
+    # use the current thread's id if possible
+    try:
+        current_thread_id = threading.current_thread().ident
+        if current_thread_id is not None:
+            return current_thread_id
+    except AttributeError:
+        pass
+
+    # if we can't get the current thread id, fall back to the main thread id
+    try:
+        main_thread_id = threading.main_thread().ident
+        if main_thread_id is not None:
+            return main_thread_id
+    except AttributeError:
+        pass
+
+    # we've tried everything, time to give up
+    return None
+
+
 class Profile(object):
     def __init__(
         self,
-        scheduler,  # type: Scheduler
         transaction,  # type: sentry_sdk.tracing.Transaction
+        hub=None,  # type: Optional[sentry_sdk.Hub]
+        scheduler=None,  # type: Optional[Scheduler]
     ):
         # type: (...) -> None
-        self.scheduler = scheduler
-        self.transaction = transaction
+        self.scheduler = _scheduler if scheduler is None else scheduler
+        self.hub = hub
+
+        self.event_id = uuid.uuid4().hex  # type: str
+
+        # Here, we assume that the sampling decision on the transaction has been finalized.
+        #
+        # We cannot keep a reference to the transaction around here because it'll create
+        # a reference cycle. So we opt to pull out just the necessary attributes.
+        self.sampled = transaction.sampled  # type: Optional[bool]
+
+        # Various framework integrations are capable of overwriting the active thread id.
+        # If it is set to `None` at the end of the profile, we fall back to the default.
+        self._default_active_thread_id = get_current_thread_id() or 0  # type: int
+        self.active_thread_id = None  # type: Optional[int]
+
         self.start_ns = 0  # type: int
         self.stop_ns = 0  # type: int
         self.active = False  # type: bool
@@ -314,27 +433,121 @@ class Profile(object):
         self.stacks = []  # type: List[ProcessedStack]
         self.samples = []  # type: List[ProcessedSample]
 
+        self.unique_samples = 0
+
         transaction._profile = self
 
-    def __enter__(self):
+    def update_active_thread_id(self):
         # type: () -> None
+        self.active_thread_id = get_current_thread_id()
+
+    def _set_initial_sampling_decision(self, sampling_context):
+        # type: (SamplingContext) -> None
+        """
+        Sets the profile's sampling decision according to the following
+        precdence rules:
+
+        1. If the transaction to be profiled is not sampled, that decision
+        will be used, regardless of anything else.
+
+        2. Use `profiles_sample_rate` to decide.
+        """
+
+        # The corresponding transaction was not sampled,
+        # so don't generate a profile for it.
+        if not self.sampled:
+            self.sampled = False
+            return
+
+        # The profiler hasn't been properly initialized.
+        if self.scheduler is None:
+            self.sampled = False
+            return
+
+        hub = self.hub or sentry_sdk.Hub.current
+        client = hub.client
+
+        # The client is None, so we can't get the sample rate.
+        if client is None:
+            self.sampled = False
+            return
+
+        options = client.options
+        sample_rate = options["_experiments"].get("profiles_sample_rate")
+
+        # The profiles_sample_rate option was not set, so profiling
+        # was never enabled.
+        if sample_rate is None:
+            self.sampled = False
+            return
+
+        # Now we roll the dice. random.random is inclusive of 0, but not of 1,
+        # so strict < is safe here. In case sample_rate is a boolean, cast it
+        # to a float (True becomes 1.0 and False becomes 0.0)
+        self.sampled = random.random() < float(sample_rate)
+
+    def get_profile_context(self):
+        # type: () -> ProfileContext
+        return {"profile_id": self.event_id}
+
+    def start(self):
+        # type: () -> None
+        if not self.sampled or self.active:
+            return
+
+        assert self.scheduler, "No scheduler specified"
+        self.active = True
         self.start_ns = nanosecond_time()
         self.scheduler.start_profiling(self)
 
-    def __exit__(self, ty, value, tb):
-        # type: (Optional[Any], Optional[Any], Optional[Any]) -> None
+    def stop(self):
+        # type: () -> None
+        if not self.sampled or not self.active:
+            return
+
+        assert self.scheduler, "No scheduler specified"
+        self.active = False
         self.scheduler.stop_profiling(self)
         self.stop_ns = nanosecond_time()
 
+    def __enter__(self):
+        # type: () -> Profile
+        hub = self.hub or sentry_sdk.Hub.current
+
+        _, scope = hub._stack[-1]
+        old_profile = scope.profile
+        scope.profile = self
+
+        self._context_manager_state = (hub, scope, old_profile)
+
+        self.start()
+
+        return self
+
+    def __exit__(self, ty, value, tb):
+        # type: (Optional[Any], Optional[Any], Optional[Any]) -> None
+        self.stop()
+
+        _, scope, old_profile = self._context_manager_state
+        del self._context_manager_state
+
+        scope.profile = old_profile
+
     def write(self, ts, sample):
         # type: (int, RawSample) -> None
+        if not self.active:
+            return
+
         if ts < self.start_ns:
             return
 
         offset = ts - self.start_ns
         if offset > MAX_PROFILE_DURATION_NS:
+            self.stop()
             return
 
+        self.unique_samples += 1
+
         elapsed_since_start_ns = str(offset)
 
         for tid, (stack_id, stack) in sample:
@@ -385,21 +598,20 @@ class Profile(object):
             "thread_metadata": thread_metadata,
         }
 
-    def to_json(self, event_opt, options, scope):
-        # type: (Any, Dict[str, Any], Optional[sentry_sdk.scope.Scope]) -> Dict[str, Any]
-
+    def to_json(self, event_opt, options):
+        # type: (Any, Dict[str, Any]) -> Dict[str, Any]
         profile = self.process()
 
         handle_in_app_impl(
-            profile["frames"], options["in_app_exclude"], options["in_app_include"]
+            profile["frames"],
+            options["in_app_exclude"],
+            options["in_app_include"],
+            default_in_app=False,  # Do not default a frame to `in_app: True`
         )
 
-        # the active thread id from the scope always take priorty if it exists
-        active_thread_id = None if scope is None else scope.active_thread_id
-
         return {
             "environment": event_opt.get("environment"),
-            "event_id": uuid.uuid4().hex,
+            "event_id": self.event_id,
             "platform": "python",
             "profile": profile,
             "release": event_opt.get("release", ""),
@@ -419,7 +631,7 @@ class Profile(object):
             "transactions": [
                 {
                     "id": event_opt["event_id"],
-                    "name": self.transaction.name,
+                    "name": event_opt["transaction"],
                     # we start the transaction before the profile and this is
                     # the transaction start time relative to the profile, so we
                     # hardcode it to 0 until we can start the profile before
@@ -427,16 +639,24 @@ class Profile(object):
                     # use the duration of the profile instead of the transaction
                     # because we end the transaction after the profile
                     "relative_end_ns": str(self.stop_ns - self.start_ns),
-                    "trace_id": self.transaction.trace_id,
+                    "trace_id": event_opt["contexts"]["trace"]["trace_id"],
                     "active_thread_id": str(
-                        self.transaction._active_thread_id
-                        if active_thread_id is None
-                        else active_thread_id
+                        self._default_active_thread_id
+                        if self.active_thread_id is None
+                        else self.active_thread_id
                     ),
                 }
             ],
         }
 
+    def valid(self):
+        # type: () -> bool
+        return (
+            self.sampled is not None
+            and self.sampled
+            and self.unique_samples >= PROFILE_MINIMUM_SAMPLES
+        )
+
 
 class Scheduler(object):
     mode = "unknown"
@@ -445,6 +665,11 @@ class Scheduler(object):
         # type: (int) -> None
         self.interval = 1.0 / frequency
 
+        self.sampler = self.make_sampler()
+
+        self.new_profiles = deque()  # type: Deque[Profile]
+        self.active_profiles = set()  # type: Set[Profile]
+
     def __enter__(self):
         # type: () -> Scheduler
         self.setup()
@@ -464,60 +689,11 @@ class Scheduler(object):
 
     def start_profiling(self, profile):
         # type: (Profile) -> None
-        raise NotImplementedError
-
-    def stop_profiling(self, profile):
-        # type: (Profile) -> None
-        raise NotImplementedError
-
-
-class ThreadScheduler(Scheduler):
-    """
-    This abstract scheduler is based on running a daemon thread that will call
-    the sampler at a regular interval.
-    """
-
-    mode = "thread"
-    name = None  # type: Optional[str]
-
-    def __init__(self, frequency):
-        # type: (int) -> None
-        super(ThreadScheduler, self).__init__(frequency=frequency)
-
-        self.sampler = self.make_sampler()
-
-        # used to signal to the thread that it should stop
-        self.event = threading.Event()
-
-        # make sure the thread is a daemon here otherwise this
-        # can keep the application running after other threads
-        # have exited
-        self.thread = threading.Thread(name=self.name, target=self.run, daemon=True)
-
-        self.new_profiles = deque()  # type: Deque[Profile]
-        self.active_profiles = set()  # type: Set[Profile]
-
-    def setup(self):
-        # type: () -> None
-        self.thread.start()
-
-    def teardown(self):
-        # type: () -> None
-        self.event.set()
-        self.thread.join()
-
-    def start_profiling(self, profile):
-        # type: (Profile) -> None
-        profile.active = True
         self.new_profiles.append(profile)
 
     def stop_profiling(self, profile):
         # type: (Profile) -> None
-        profile.active = False
-
-    def run(self):
-        # type: () -> None
-        raise NotImplementedError
+        pass
 
     def make_sampler(self):
         # type: () -> Callable[..., None]
@@ -600,23 +776,42 @@ class ThreadScheduler(Scheduler):
         return _sample_stack
 
 
-class SleepScheduler(ThreadScheduler):
+class ThreadScheduler(Scheduler):
     """
-    This scheduler uses time.sleep to wait the required interval before calling
-    the sampling function.
+    This scheduler is based on running a daemon thread that will call
+    the sampler at a regular interval.
     """
 
-    mode = "sleep"
-    name = "sentry.profiler.SleepScheduler"
+    mode = "thread"
+    name = "sentry.profiler.ThreadScheduler"
+
+    def __init__(self, frequency):
+        # type: (int) -> None
+        super(ThreadScheduler, self).__init__(frequency=frequency)
+
+        # used to signal to the thread that it should stop
+        self.running = False
+
+        # make sure the thread is a daemon here otherwise this
+        # can keep the application running after other threads
+        # have exited
+        self.thread = threading.Thread(name=self.name, target=self.run, daemon=True)
+
+    def setup(self):
+        # type: () -> None
+        self.running = True
+        self.thread.start()
+
+    def teardown(self):
+        # type: () -> None
+        self.running = False
+        self.thread.join()
 
     def run(self):
         # type: () -> None
         last = time.perf_counter()
 
-        while True:
-            if self.event.is_set():
-                break
-
+        while self.running:
             self.sampler()
 
             # some time may have elapsed since the last time
@@ -624,51 +819,70 @@ class SleepScheduler(ThreadScheduler):
             # not sleep for too long
             elapsed = time.perf_counter() - last
             if elapsed < self.interval:
-                time.sleep(self.interval - elapsed)
+                thread_sleep(self.interval - elapsed)
 
             # after sleeping, make sure to take the current
             # timestamp so we can use it next iteration
             last = time.perf_counter()
 
 
-def _should_profile(transaction, hub):
-    # type: (sentry_sdk.tracing.Transaction, sentry_sdk.Hub) -> bool
+class GeventScheduler(Scheduler):
+    """
+    This scheduler is based on the thread scheduler but adapted to work with
+    gevent. When using gevent, it may monkey patch the threading modules
+    (`threading` and `_thread`). This results in the use of greenlets instead
+    of native threads.
 
-    # The corresponding transaction was not sampled,
-    # so don't generate a profile for it.
-    if not transaction.sampled:
-        return False
+    This is an issue because the sampler CANNOT run in a greenlet because
+    1. Other greenlets doing sync work will prevent the sampler from running
+    2. The greenlet runs in the same thread as other greenlets so when taking
+       a sample, other greenlets will have been evicted from the thread. This
+       results in a sample containing only the sampler's code.
+    """
 
-    # The profiler hasn't been properly initialized.
-    if _scheduler is None:
-        return False
+    mode = "gevent"
+    name = "sentry.profiler.GeventScheduler"
 
-    client = hub.client
+    def __init__(self, frequency):
+        # type: (int) -> None
 
-    # The client is None, so we can't get the sample rate.
-    if client is None:
-        return False
+        # This can throw an ImportError that must be caught if `gevent` is
+        # not installed.
+        from gevent.threadpool import ThreadPool  # type: ignore
 
-    options = client.options
-    profiles_sample_rate = options["_experiments"].get("profiles_sample_rate")
+        super(GeventScheduler, self).__init__(frequency=frequency)
 
-    # The profiles_sample_rate option was not set, so profiling
-    # was never enabled.
-    if profiles_sample_rate is None:
-        return False
+        # used to signal to the thread that it should stop
+        self.running = False
 
-    return random.random() < float(profiles_sample_rate)
+        # Using gevent's ThreadPool allows us to bypass greenlets and spawn
+        # native threads.
+        self.pool = ThreadPool(1)
 
+    def setup(self):
+        # type: () -> None
+        self.running = True
+        self.pool.spawn(self.run)
 
-@contextmanager
-def start_profiling(transaction, hub=None):
-    # type: (sentry_sdk.tracing.Transaction, Optional[sentry_sdk.Hub]) -> Generator[None, None, None]
-    hub = hub or sentry_sdk.Hub.current
+    def teardown(self):
+        # type: () -> None
+        self.running = False
+        self.pool.join()
 
-    # if profiling was not enabled, this should be a noop
-    if _should_profile(transaction, hub):
-        assert _scheduler is not None
-        with Profile(_scheduler, transaction):
-            yield
-    else:
-        yield
+    def run(self):
+        # type: () -> None
+        last = time.perf_counter()
+
+        while self.running:
+            self.sampler()
+
+            # some time may have elapsed since the last time
+            # we sampled, so we need to account for that and
+            # not sleep for too long
+            elapsed = time.perf_counter() - last
+            if elapsed < self.interval:
+                thread_sleep(self.interval - elapsed)
+
+            # after sleeping, make sure to take the current
+            # timestamp so we can use it next iteration
+            last = time.perf_counter()
